// Copyright 2000-2007 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package org.ibex.io;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;
import org.ibex.util.*;
import org.ibex.crypto.*;

// Features:
//   - automatically flush writer before reading on linked read/write pairs
//   - no checked exceptions thrown
//   - unified write(char), print(char), and write(byte)
//   - unreading/peeking
//   - transcribe
//   - append (daisy-chaining)
//   - automatically closes input streams when end reached

/** plays the role of InputStream, OutputStream, Reader and Writer, with logging and unchecked exceptions */
public class Stream {

    protected final In in;
    protected final Out out;
    private         String newLine = "\r\n";

    public  Stream(byte[] b, int off, int len)       { this.in = new Stream.In(new ByteArrayInputStream(b, off, len)); this.out=null; }
    public  Stream(InputStream in)                   { this.in = new Stream.In(in); this.out = null; }
    public  Stream(                OutputStream out) { this.in = null;              this.out = new Stream.Out(out); }
    public  Stream(InputStream in, OutputStream out) { this.in = new Stream.In(in); this.out = new Stream.Out(out); }
    public  Stream(String s)                         { this(new ByteArrayInputStream(s.getBytes())); }
    public  Stream(File f)                           {
        try { this.in = new Stream.In(new FileInputStream(f)); } catch (IOException e) { ioe(e); throw new Error(); }
        this.out = null;
    }
    public  Stream(Socket s) {
        try { this.in = new Stream.In(s.getInputStream());    } catch (IOException e) { ioe(e); throw new Error(); }
        try { this.out = new Stream.Out(s.getOutputStream()); } catch (IOException e) { ioe(e); throw new Error(); }
    }

    // Main API //////////////////////////////////////////////////////////////////////////////

    public char   peekc()                          { flush(); return in.getc(true); }
    public char   getc()                           { flush(); return in.getc(false); }
    public String readln()                         { flush(); return in.readln(); }
    public void   print(String s)                  { out.write(s); flush(); }
    public void   println(String s)                { print(s); print(newLine); }
    public void   flush()                          { if (out != null) out.flushWriter(); }
    public void writeBytes(byte[] b, int off, int len) { try { out.write(b, off, len); } catch (IOException e) { ioe(e); } }
    public int    read(byte[] b, int off, int len) { flush(); return in.readBytes(b, off, len); }
    public int    read(char[] c, int off, int len) { flush(); return in.readChars(c, off, len); }
    public void   close()                          { try { if (in!=null) in.close(); } finally { if (out!=null) out.close(); } }
    public void   setNewline(String s)             { newLine = s; }
    public InputStream getInputStream() { return in; }

    public void   setInputDigest(Digest d) { in.bbis.digest = d; }

    private static class Out extends BufferedOutputStream {
        private Writer writer = new OutputStreamWriter(this);
        public Out(OutputStream out) { super(out); }
        public  void close() { try { super.close(); } catch (Exception e) { Log.error(this, e); } }
        public  void write(String s) { try { writer.write(s); } catch (IOException e) { ioe(e); } }
        private void flushWriter() { try { writer.flush(); } catch (IOException e) { ioe(e); } }
        private boolean flushing = false;
        public  void flush() {
            if (flushing) return;
            try {
                flushing = true;
                try {
                    writer.flush();
                } finally { flushing = false; }
                super.flush();
            } catch (IOException e) { ioe(e); }
        }
    }

    private class In extends InputStream {
        private ByteBufInputStream bbis;
        private CharBufReader cbr;
        public  Reader reader;
        private Writer unreader;

        public char getc(boolean peek) { return cbr.getc(peek); }
        public String readln() { return cbr.readln(); }
        public int read() { return bbis.read(); }
        public int read(byte[] b) { try { return bbis.read(b); } catch (IOException e) { ioe(e); return 0; } }
        public int read(byte[] b, int off, int len) { return bbis.read(b, off, len); }
        public void close() { try { cbr.close(); } catch (Exception e) { Log.error(this, e); } }
        public int readBytes(byte[] b, int off, int len) { return bbis.read(b, off, len); }
        public int readChars(char[] c, int off, int len) { return cbr.read(c, off, len); }

        public In(InputStream in) {
            bbis = new ByteBufInputStream(in) {
                    public void preread() {
                        cbr.unbuffer(unreader);
                        try {
                            if (!cbr.ready()) return;
                        } catch (IOException e) { ioe(e); }
                        char[] c = new char[20];
                        while(true) {
                            try {
                                if (!cbr.ready()) break;
                                int numread = cbr.read(c, 0, c.length);
                                if (numread == -1) break;
                                unreader.write(c, 0, numread);
                            } catch (IOException e) { ioe(e); }
                        }
                    }
                };
            unreader = new OutputStreamWriter(new UnReaderStream(bbis));
            cbr = new CharBufReader(new InputStreamReader(bbis));
        }
    }

    // Utilities: append() and transcribe() ///////////////////////////////////////////////////////

    public Stream append(String in_next) { return appendStream(new Stream(in_next)); }
    public Stream appendStream(Stream in_next) { in.bbis.appendStream(in_next); return this; }

    public void transcribe(Stream out) { transcribe(out, false); }
    public void transcribe(Stream out, boolean close) {
        byte[] buf = new byte[1024];
        while(true) {
            int numread = in.read(buf, 0, buf.length);
            if (numread==-1) { in.close(); break; }
            out.writeBytes(buf, 0, numread);
        }
        if (close) out.close();
    }

    public void transcribe(StringBuffer out) {
        char[] buf = new char[1024];
        while(true) {
            int numread = in.readChars(buf, 0, buf.length);
            if (numread==-1) { in.close(); return; }
            out.append(buf, 0, numread);
        }
    }

    public static int countLines(Stream s) {
        int ret = 0;
        while(s.readln() != null) ret++;
        s.close();
        return ret;
    }

    // FIXME: ugly
    public static int countBytes(Stream s) {
        int ret = 0;
        while(s.in.read() != -1) ret++;
        s.close();
        return ret;
    }

    // Exceptions //////////////////////////////////////////////////////////////////////////////

    static int ioe(IOException e) {
        if (e instanceof SocketException && e.toString().indexOf("Connection reset")!=-1)
            throw new Closed(e.getMessage());
        throw new StreamException(e);
    }
    public static class StreamException extends RuntimeException {
        public StreamException(Exception e) { super(e); }
        public StreamException(String s)    { super(s); }
    }
    public static class EOF             extends StreamException  { public EOF() { super("End of stream"); } }
    public static class Closed          extends StreamException  { public Closed(String s) { super(s); } }
}
